from bitwarden_import_msecure.__about__ import __version__
from bitwarden_import_msecure.main import bitwarden_import_msecure


def test_version_option(runner):
    result = runner.invoke(bitwarden_import_msecure, ["--version"])
    assert result.exit_code == 0
    assert __version__ in result.output


def test_force_and_patch_together(runner, tmp_path):
    dummy_input = tmp_path / "dummy_input.csv"
    dummy_input.touch()  # Create an empty file to satisfy existence check

    dummy_output = tmp_path / "dummy_output.json"
    dummy_output.touch()  # Create an empty file to potentially be patched

    result = runner.invoke(
        bitwarden_import_msecure, [str(dummy_input), str(dummy_output), "--force", "--patch"]
    )

    assert result.exit_code == 1
    assert "--force and --patch cannot be used simultaneously." in result.output


def test_patch_without_existing_output_file(runner, tmp_path):
    dummy_input = tmp_path / "dummy_input.csv"
    dummy_input.touch()  # Create an empty file to satisfy the input file existence check

    non_existent_output = tmp_path / "non_existent_output.json"

    result = runner.invoke(
        bitwarden_import_msecure, [str(dummy_input), str(non_existent_output), "--patch"]
    )

    assert result.exit_code == 1
    assert f"Output file `{non_existent_output}` does not exist." in result.output.replace("\n", "")


def test_patch_with_incorrect_format(runner, tmp_path):
    dummy_input = tmp_path / "dummy_input.csv"
    dummy_input.touch()  # Create an empty file

    incorrect_format_output = tmp_path / "output.csv"
    incorrect_format_output.touch()  # Ensure the file exists for the patch command

    result = runner.invoke(
        bitwarden_import_msecure,
        [str(dummy_input), str(incorrect_format_output), "--format", "csv", "--patch"],
    )

    assert result.exit_code == 1
    assert "Patching is only supported for JSON format." in result.output


def test_overwrite_without_force(runner, tmp_path):
    # Create a dummy input file to satisfy the input file check
    dummy_input = tmp_path / "dummy_input.csv"
    dummy_input.touch()  # Create an empty file

    # Create a file to attempt to overwrite
    test_output = tmp_path / "test_output.json"
    test_output.write_text("test content")  # create and write to the file

    # Run the command without --force, correctly setting input and output file arguments
    result = runner.invoke(
        bitwarden_import_msecure,
        [
            str(dummy_input),  # the input file
            str(test_output),  # the output file
            "--format",
            "json",  # specifying format explicitly if necessary
        ],
    )

    assert result.exit_code == 1, f"Expected exit code 1, got {result.exit_code}"
    assert "--force" in result.output.replace("\n", ""), (
        f"Expected error message not found in output: {result.output}"
    )



"""Bitwarden Import mSecure Export."""

from pathlib import Path

import rich_click as click
from rich.console import Console

from bitwarden_import_msecure import msecure_to_bitwarden
from bitwarden_import_msecure.__about__ import __version__

click.rich_click.USE_MARKDOWN = True


OUTPUT_FILE_DEFAULT = "bitwarden"
NOTES_MODE = "notes"


def error(message: str, show_patch_help: bool = True) -> None:
    """Print error message and exit."""
    console = Console()
    console.print(message, style="bold red")
    if show_patch_help:
        msecure_to_bitwarden.patch_help()
    raise click.Abort()


@click.command()
@click.version_option(version=__version__, prog_name="bitwarden-import-msecure")
@click.argument("input_file", type=click.Path(exists=True), required=False)
@click.argument("output_file", type=click.Path(), required=False)
@click.option("--force", is_flag=True, help="Overwrite the output file if it exists.")
@click.option(
    "--patch",
    is_flag=True,
    help="Patch Bitwarden export with data from mSecure export that previously was not imported. "
    "See `--patch-help` for more details.",
)
@click.option("--patch-help", is_flag=True, help="Show help for `--patch` option.")
@click.option(
    "--extra-fields",
    type=click.Choice(["custom-fields", NOTES_MODE]),
    default="custom-fields",
    help=(
        "How to handle mSecure fields that don't match Bitwarden fields."
        f"By default, they are added as custom fields. Use '{NOTES_MODE}' to add them to notes."
    ),
)
@click.option(
    "--format",
    "output_format",
    type=click.Choice(["csv", "json"]),
    default="json",
    help="Output file format. JSON by default. CSV is legacy format with less features.",
)
def bitwarden_import_msecure(  # noqa: PLR0913
    input_file: str,
    output_file: str,
    force: bool,
    patch: bool,
    extra_fields: str,
    output_format: str,
    patch_help: bool,
) -> None:
    """
    Converts file `INPUT_FILE` exported from mSecure to Bitwarden compatible format
    to `OUTPUT_FILE`.

    - Export CSV from mSecure
    - Run this script on the exported CSV file
    - Import the processed file into Bitwarden
    """
    if patch_help:
        msecure_to_bitwarden.patch_help()
        return

    if not input_file:
        error("No input file provided.")
    input_path = Path(input_file)
    if not input_path.exists():
        error(f"Input file `{input_path}` does not exist.")

    if force and patch:
        error("--force and --patch cannot be used simultaneously.")

    output_path = (
        Path(output_file)
        if output_file
        else input_path.parent / f"{OUTPUT_FILE_DEFAULT}.{output_format}"
    )

    if patch:
        if output_format != "json":
            error("Patching is only supported for JSON format.", show_patch_help=True)
        if not output_path.exists():
            error(
                f"Output file `{output_path}` does not exist. Cannot patch un-existed file.",
                show_patch_help=True,
            )
        msecure_to_bitwarden.patch(input_path, output_path)
    else:
        if output_path.exists() and not force:
            error(f"Output file `{output_path}` already exists. Use --force to overwrite.")

        msecure_to_bitwarden.convert(
            input_path,
            output_path,
            output_format=output_format,
            extra_fields_to_notes=extra_fields == NOTES_MODE,
        )
    click.echo(f"File to import into Bitwarden saved to `{output_path}`")


if __name__ == "__main__":  # pragma: no cover
    bitwarden_import_msecure()  # pylint: disable=no-value-for-parameter



"""Conversion logic."""

import csv
import json
from functools import reduce
from pathlib import Path
from typing import Optional

import rich_click as click
from rich.console import Console
from rich.markdown import Markdown
from rich.panel import Panel
from rich.theme import Theme

from bitwarden_import_msecure.bitwarden_csv import BitwardenCsv
from bitwarden_import_msecure.bitwarden_json import BitwardenJson
from bitwarden_import_msecure.msecure import import_msecure_row


def patch(input_path: Path, output_path: Path) -> None:  # noqa: C901
    """Patch Bitwarden export with data from mSecure export that previously was not imported

    Some old versions of `bitwarden-import-msecure` worked incorrectly.
    For example versions before 1.5.0 did not export logins' URLs.

    If you migrated to Bitwarden some time ago and cannot just drop all records
    and import them again, use option `--patch`:

    - export json from Bitwarden, let's name the result as `bitwarden_new.json`.
    please backup this file in case something goes wrong
    - patch this export with data from the mSecure export
        `bitwarden-import-msecure "mSecure Export File.csv" bitwarden_new.json --patch`
    - now you have `bitwarden_new.json` with additional data
    - unfortunately Bitwarden does not respect item IDs on import, so to avoid duplicates
    remove all items from Bitwarden, preferably using web interface.
    It is save as you have full backup in `bitwarden_new.json`
    - import bitwarden_new.json to Bitwarden as Bitwarden json file
    - clean up mSecure export file, `bitwarden_new.json` and it's backup
    """

    def get_row_dict(csv_row: list[str]) -> Optional[dict[str, str]]:
        """Get dict from mSecure CSV line."""
        if csv_row and not csv_row[0].startswith("mSecure"):
            row = import_msecure_row(csv_row, False)
            if row and row["type"] == "login":
                return {"type": "login", "login_uri": row["login_uri"], "name": row["name"]}
        return None

    def filter_logins_with_url(rows: list[dict[str, str]]) -> list[dict[str, str]]:
        """Filter logins with non-empty URL."""
        return list(
            filter(lambda x: x is not None and x["type"] == "login" and x["login_uri"], rows),
        )

    def reduce_to_uri_dict(rows: list[dict[str, str]]) -> dict[str, str]:
        """Reduce list of logins' dicts to dict with login name as key and login URL as value."""

        def reduce_reporting_name_collisions(
            acc: dict[str, str],
            row: dict[str, str],
        ) -> dict[str, str]:
            if row["name"] in acc and acc[row["name"]] != row["login_uri"]:
                print(
                    f"Name collision: item `{row['name']}`, has different URLs: "
                    f"`{acc[row['name']]}` and `{row['login_uri']}`.\n"
                    f"Using first one.",
                )
            else:
                acc[row["name"]] = row["login_uri"]
            return acc

        return reduce(reduce_reporting_name_collisions, rows, {})

    if not output_path.exists():
        click.echo(f"Output file `{output_path}` does not exist.")
        raise click.Abort()

    try:
        print(f"Reading output file: {output_path}..")
        with output_path.open("r+") as file:
            output_data = json.load(file)

            with input_path.open(newline="", encoding="utf-8") as infile:
                reader = csv.reader(infile, delimiter=",")
                rows = [row for row in map(get_row_dict, reader) if row is not None]
                uri_dict = reduce_to_uri_dict(filter_logins_with_url(rows))

            replaced = 0
            for item in output_data.get("items", []):
                if item.get("type") == 1 and (
                    item["name"] in uri_dict and not item.get("login", {}).get("uris", [])
                ):
                    item["login"]["uris"] = [{"match": None, "uri": uri_dict[item["name"]]}]
                    replaced += 1
            click.echo(f"Added {replaced} URLs.")

            file.seek(0)
            json.dump(output_data, file, indent=4)
            file.truncate()

    except json.JSONDecodeError as e:
        print(f"Error: {output_path} is not a valid JSON file:\n{e}")
        return
    except FileNotFoundError as e:
        print(f"Error: {output_path} not found:\n{e}")
        return


def patch_help() -> None:
    """Show help message for `--patch` option."""
    custom_theme = Theme(
        {
            "markdown.heading": "bold magenta",  # Styling for headings
            "markdown.code": "bold",  # Code blocks often stand out
            "markdown.list": "dim",  # Lists are usually less emphasized
            "markdown.block_quote": "italic",  # Block quotes may be italicized
            "markdown.link": "underline blue",  # Links can be underlined and blue
            "markdown.italic": "italic",  # Explicit style for italic text
            "markdown.bold": "bold",  # Explicit style for bold text
        },
    )
    console = Console(theme=custom_theme)
    assert patch.__doc__
    lines = [line.strip() for line in patch.__doc__.strip().split("\n")]
    title = lines[0]
    markdown_content = "\n".join(lines[2:])  # Skip title and empty line
    markdown = Markdown(markdown_content)
    panel = Panel(markdown, title=title, border_style="gray46")

    console.print(panel)


def convert(
    input_path: Path,
    output_path: Path,
    *,
    output_format: str,
    extra_fields_to_notes: bool,
) -> None:
    """Convert mSecure export to Bitwarden format."""
    writer = BitwardenCsv(output_path) if output_format == "csv" else BitwardenJson(output_path)
    with input_path.open(newline="", encoding="utf-8") as infile:
        reader = csv.reader(infile, delimiter=",")
        for row in reader:
            if row and not row[0].startswith("mSecure"):
                data = import_msecure_row(row, extra_fields_to_notes)
                writer.write_record(data)
    writer.close()
